#include "geotraceman.h"

/*
Reads in the cil map specified by the the map name passed. It checks to see if there was an error
reading the file.
*/
int read_cil_map()
{
	int error;
	int side_size;
	int region_range;
	char map_name[File_Name_Length];

	strcpy(map_name, Map_Directory);
	strcat(map_name, Cil_Map_File_Names[ms.detail]);

	error = read_map(&cil_map, map_name, &side_size, &region_range);

	if (error != 0) {
		return(1);
	}

	ms.side_size = side_size;
	ms.region_range = region_range;
	
	ms.num_regions = ms.side_size * ms.side_size;
	ms.max_range = ( (long) ms.side_size * (long) ms.region_range);
	
	return(0);

}

/*
Reads in a pdy or riv map specified by the map name passed. It checks to see if there was an
error reading the file, or if the file is inconsistent with the current cil file.
*/ 
int read_other_map(map_type)
	char *map_type;
{
	int side_size;
	int region_range;
	int error;
	Region_Struct *map;
	char map_name[File_Name_Length];

	strcpy(map_name, Map_Directory);

	if (strcmp(map_type,"riv") == NULL)
		strcat(map_name, Riv_Map_File_Names[ms.detail]);
	else 
		strcat(map_name, Pdy_Map_File_Names[ms.detail]);

	error = read_map(&map, map_name, &side_size, &region_range);

	if (error != 0) {
		return(1);
	}

	if (side_size != ms.side_size) {
		return(1);
	}

	if (region_range != ms.region_range) {
		return(1);
	}

	if (strcmp(map_type,"riv") == NULL)
		riv_map = map;
	else 
		pdy_map = map;

	return(0);
}

/*
Reads in the map specified by the map name passed as an argument, and passes back the
side_size of the map, plus the region_range, max_range and total number of regions.
*/
int read_map(regs_ptr, map_name, side_size, region_range)
	Region_Struct **regs_ptr;
	char *map_name;
	int *side_size;
	int *region_range;
{
	FILE *mapfp; /* File pointer for the map file */
	int n,j; /* Loop counters */
	int num_regions;/* Total number of regions in the map */
	int memory_size; /* The amount of memory to malloc */
	Region_Struct *regs; /* A direct pointer to the region structures */

	/* Open the file and check the result */
	mapfp = fopen(map_name, "r");
	if (mapfp == NULL) {
		return(1);
	}

	/* Input the side size, and the range per region of the map */
	fread(side_size, sizeof(int), 1, mapfp);
	*side_size = htonl(*side_size);
	fread(region_range, sizeof(int), 1, mapfp);
	*region_range = htonl(*region_range);

	/* Calculate the total number of regions in the map */
	num_regions = (*side_size) * (*side_size);
	
	/* Allocate the space for the region structures */
	(*regs_ptr) = (Region_Struct *) malloc(num_regions * sizeof(Region_Struct));
	if ((*regs_ptr) == NULL) {
		return(1);
	}

	/* Make regs equal to what regs_ptr is pointing to */
	regs = *regs_ptr;

	/* Input each region from the file */
	for(n=0; n < num_regions; n++) {
		/* Input the number of vectors in a region */
		fread(&regs[n].nvect, sizeof(int), 1, mapfp);
		regs[n].nvect = htonl(regs[n].nvect);

		/* Allocate space for the segment breakdown per vector, and and an array of
		   pointers to the actual vectors data if the number of vectors in the region
		   is greater than zero */

		if (regs[n].nvect != 0) {
			memory_size = regs[n].nvect * sizeof(int);
			regs[n].nseg = (int *) malloc(memory_size);
			if (regs[n].nseg == NULL) {
				return(1);
			}
	
			memory_size = regs[n].nvect * sizeof(unsigned char *);
			regs[n].line = (unsigned char **) malloc(memory_size);
			if (regs[n].line == NULL) {
				return(1);
			}

			/* Get the segment breakdown per vector */ 
			fread(regs[n].nseg, sizeof(int), regs[n].nvect, mapfp);
	
			/* Read in the points themselves */
			for(j=0; j < regs[n].nvect; j++) {
				regs[n].nseg[j] = htonl(regs[n].nseg[j]);
				if(regs[n].nseg[j] != 0) {	
					memory_size = regs[n].nseg[j] * 2 * sizeof(unsigned char);
					regs[n].line[j] = (unsigned char *) malloc(memory_size);
					fread(regs[n].line[j], sizeof(unsigned char), (regs[n].nseg[j] * 2), mapfp);
				}

			}	
		}
	}	 
	fclose(mapfp);
}

/*
Frees the area of memory held by a map
*/
void
free_regions(regs_ptr, num_regions)
	Region_Struct **regs_ptr;
	int num_regions;
{
	int i, j; /* Counter variables */

	for(i=0; i < num_regions; i++) {
		for(j=0; j < regs_ptr[0][i].nvect; j++)
			free((regs_ptr[0][i].line[j]));
		free((regs_ptr[0][i].nseg));
		free(&(regs_ptr[0][i]));
	}

	free(*regs_ptr);	
}

/*
Releases the memory held for the old maps, and reads in the new ones
*/
int swap_maps()
{
	/* Release the three maps */
	free_regions(&cil_map, ms.num_regions);	
	free_regions(&riv_map, ms.num_regions);	
	free_regions(&pdy_map, ms.num_regions);	

	/* Read in the three new maps */
	if(read_cil_map() != 0) {
		fprintf(stderr,"GeoTraceMan: Error reading cil map\n");
		exit(1);
	}

	if(read_other_map("riv") != 0) {
		fprintf(stderr,"GeoTraceMan: Error reading riv map\n");
		exit(1);
	}

	if(read_other_map("pdy") != 0) {
		fprintf(stderr,"GeoTraceMan: Error reading pby map\n");
		exit(1);
	}
}
