#include "geotraceman.h"

/***
Functions to do with dumping the display to a ps file
***/

#define			my_mask		707
#define out(arg)        fprintf(psf,"%s\n",arg)

Boolean convert_values();
FILE		*psf;
float		trace_scale_x;
float		trace_scale_y;


float deg_to_rad_ps(degree)
float degree;
{
    return degree * PI / 180.0;
}

void
post_scale_font()
{
	float	xfont;

	xfont = (9.0 +  ms.detail) / 2.0;
	fprintf(psf,"%3.2f scalefont\n",xfont);
}
void
post_move (x, y)
float  x, y;
{
    float   scaledx, scaledy;

    scaledx = (x - ms.x.min_vis) / trace_scale_x ;
    scaledy = (y - ms.y.min_vis) / trace_scale_y ;
    fprintf (psf, "%3.2f %3.2f M\n", scaledx, scaledy);
}

post_draw (x, y)
float  x, y;
{
    float   scaledx, scaledy;

    scaledx = (x - ms.x.min_vis) / trace_scale_x ;
    scaledy = (y - ms.y.min_vis) / trace_scale_y ;
    fprintf (psf, "%3.2f %3.2f L\n", scaledx, scaledy);
}

post_trans (x, y)
float  x, y;
{
    float   scaledx, scaledy;

    scaledx = (x - ms.x.min_vis) / trace_scale_x ;
    scaledy = (y - ms.y.min_vis) / trace_scale_y ;
    fprintf (psf, "%3.2f %3.2f translate\n", scaledx, scaledy);
}

pscharstr (str)
char   *str;
{
    fprintf (psf, "(%s) H\n", str);
}


post_arc (x, y, radius)
float  x, y, radius;
{
    float   scaledx, scaledy, scaledrad;

    scaledx = (x - ms.x.min_vis) / trace_scale_x ;
    scaledy = (y - ms.y.min_vis) / trace_scale_y ;
    scaledrad = radius / trace_scale_x ;
    fprintf(psf,"%3.2f %3.2f %3.2f 0 360 arc\n",scaledx, scaledy, scaledrad);
}

void
ps_setrgbcolor (r, g, b)
int  r, g, b;
{
        fprintf (psf, "%d %d %d setrgbcolor\n", r, g, b);
}


void 
set_scaling(x_scale,y_scale)
	float	*x_scale,*y_scale;
{
	/* set the scaling for the postscript dump */
	if (nm_page_type = A4)
	{
		if (((ms.includes & INCL_3D) == INCL_3D) || (ms.zoom > 1))
		{
			*y_scale = (ms.y.max_vis - ms.y.min_vis)/595.27559;	
			*x_scale = (ms.x.max_vis - ms.x.min_vis)/595.27559;	
		}
		else
		{
			*y_scale = (ms.y.max_vis - ms.y.min_vis)/595.27559;	
			*x_scale = (ms.x.max_vis - ms.x.min_vis)/765.35433;
		}
	}
}

void
output_header()
{
	/* output the required header info to the PS file*/
	umask(my_mask);	
	psf=fopen(nm_postscript_output_file,"r");
	if (psf == NULL)
	{	
		/* set up macros */
		psf=fopen(nm_postscript_output_file,"w");
		fprintf(psf,"%%!PS-Abode-3.0 EPSF-3.0\n");
		fprintf(psf,"%%BoundingBox: 72 180 536 650\n");
		fprintf(psf,"/S{stroke}def\n");
		fprintf(psf,"/M{moveto}def\n");
		fprintf(psf,"/L{lineto}def\n");
		fprintf(psf,"/N{newpath}def\n");
		fprintf(psf,"/SP{showpage}def\n");
		fprintf(psf,"/H{show}def\n");
		fprintf(psf,"0 setlinewidth\n");
	}
	else
	{
		psf=fopen(nm_postscript_output_file,"a+");
	}
	/* lanscape for normal map and square portraite for world map */

	if (page_begin == 1)
	{
		page_begin = 0;
	
		if ((ms.includes & INCL_3D) == INCL_3D)
		{
			/*out("0 200 translate");*/
			out("N");
			ps_setrgbcolor(0, 0, 0); /*black*/
			out("297.92875 297.92875 223.44656 0 360 arc\n");
			out("S");
		}
		
		else
		{
			out("581.10 0 translate");
			out("90 rotate");
			out("N");
		}
	}
	switch(ps_color)
	{
		case 0:
			ps_setrgbcolor(255, 0, 255); /*purple*/
			break;
		case 1:
			ps_setrgbcolor(0, 0, 255);   /*blue*/
			break;
		case 2:
			ps_setrgbcolor(0, 255, 0);  /*green*/
			break;
		case 3:
			ps_setrgbcolor(0, 0, 0);   /*black*/
			break;
	}
	fflush(psf);
	close(psf);
	
}


/*
Outputs a particular maps regions to the ps file
*/
void 
ps_dump_some_regions(regs, min_x_reg, max_x_reg)
	Region_Struct *regs;
	int min_x_reg, max_x_reg;
{
	
	
	int        	x,y,i,j;
    	unsigned	char *line;
    	long		values[2]; 
	Region_Struct 	*reg;
	long 		leftregbdr;
	long 		leftregbdr2;
	long 		bottomregbdr;
	Boolean 	drawtwice;
	Boolean 	first_point;	
	int		tx;
	int		ty;




	set_scaling(&trace_scale_x,&trace_scale_y);
	output_header();
	psf=fopen(nm_postscript_output_file,"a+");

	if (page_flag == 1)
        {
                out("SP");
		fflush(psf);
		close(psf);
		return;
        }




    	for(x = min_x_reg; x <= max_x_reg; x++) {
		drawtwice = FALSE;
		leftregbdr = ((x + ms.x_reg_offset) * ms.region_range) + (long) ms.x_offset;

		if (leftregbdr  > ((ms.side_size - 1) * ms.region_range)) {
			if (leftregbdr < ms.x.max_vis) {
				leftregbdr2 = leftregbdr;
				drawtwice = TRUE;
			}
			leftregbdr -= ms.max_range;
		}
		else if (leftregbdr < 0) {
			if ((leftregbdr + ms.region_range) > 0) {
				leftregbdr2 = leftregbdr;
				drawtwice = TRUE;
			}

			leftregbdr += ms.max_range;
		}

		for(y = ms.y.min_vis_reg; y <= ms.y.max_vis_reg; y++) {	
			bottomregbdr = (long) (y * ms.region_range);
			reg = &regs[(x * ms.side_size) + y];
    			for (i = 0; i < reg->nvect; ++i) {
        			line = reg->line[i];
					first_point = TRUE;
        			for (j = 0; j < reg->nseg[i]; ++j) {
					if (j != 0)
	               				line += 2;
					values[0] = leftregbdr + ((long) (*line));
	       			values[1] = bottomregbdr + ((long) (*(line + 1))); 
					
					
					if ((ms.includes & INCL_3D) == INCL_3D)
					{
						if (convert_values(values) == True)
						{
							if ((values[0] <= ms.x.max_vis) && (values[1] <= ms.y.max_vis) && 
								 (values[0] > ms.x.min_vis) && (values[1] > ms.y.min_vis))
							{
								if (first_point == TRUE)
								{
									first_point = FALSE;
									post_move( ((float) values[0]) , ((float) values[1]));
								}
								else
									post_draw( ((float) values[0]) , ((float) values[1]));
									
							}
						}
					}
					else
					{
							if ((values[0] <= ms.x.max_vis) && (values[1] <= ms.y.max_vis) &&
								 (values[0] > ms.x.min_vis) && (values[1] > ms.y.min_vis))
							{
								if (first_point == TRUE)
								{
									first_point = FALSE;
									post_move( ((float) values[0]) , ((float) values[1]));
								}
								else
									post_draw( ((float) values[0]) , ((float) values[1]));
							}
					}

				}
    			}
		}

		if (drawtwice) {
			for(y = ms.y.min_vis_reg; y <= ms.y.max_vis_reg; y++) {	
				reg = &regs[(x * ms.side_size) + y];
    				for (i = 0; i < reg->nvect; ++i) {
        				line = reg->line[i];
					first_point = TRUE;
        				for (j = 0; j < reg->nseg[i]; ++j) {
						if (j != 0)
               						line += 2;
						values[0] = leftregbdr2 + (*line);
	       					values[1] = (long) (((int) (*(line + 1))) + (y * ms.region_range)); 

						if ((ms.includes & INCL_3D) == INCL_3D)
						{
							if (convert_values(values) == True)

								if ((values[0] <= ms.x.max_vis) && (values[1] <= ms.y.max_vis) &&
								 	 (values[0] > ms.x.min_vis) && (values[1] > ms.y.min_vis))
								{

									if (first_point == TRUE)
									{
										first_point = FALSE;
										post_move( ((float) values[0]) , ((float) values[1]));
									}
									else
										post_draw( ((float) values[0]) , ((float) values[1]));
								}
						}
						else
						{
							if ((values[0] <= ms.x.max_vis) && (values[1] <= ms.y.max_vis) &&
								 (values[0] > ms.x.min_vis) && (values[1] > ms.y.min_vis))
							{
								if (first_point == TRUE)
								{
									first_point = FALSE;
									post_move( ((float) values[0]) , ((float) values[1]));
								}
								else
									post_draw( ((float) values[0]) , ((float) values[1]));
							}
						}
					}
				}
			}
    		}	
	}
	out("S");
	fflush(psf);
	close(psf);
}


void
ps_dump_regions(regs) 
	Region_Struct *regs;
{
	if (ms.x.min_vis_reg <= ms.x.max_vis_reg) 
	{

		ps_dump_some_regions(regs, ms.x.min_vis_reg, ms.x.max_vis_reg);
	}
	else 
	{
		ps_dump_some_regions(regs, ms.x.min_vis_reg, (ms.side_size- 1));
		ps_dump_some_regions(regs, 0, ms.x.max_vis_reg);
	}
}	

/*****************************************************************************
the next section deals with display all the cities to the PS file
*****************************************************************************/

/*
Displays all of the cities in the cities file to the ps file 
*/
void
cities_display_ps()
{
	FILE *cf;
	float coord[2];
	struct city_struct city;
	int zoom_level;
	float xfont, yfont;
	int i;
	char filename[80];
	pcity_list head, next, prev;

	if ((ms.includes & INCL_CIT) != INCL_CIT)
		return;

	head = next = prev = NULL;

	if ((int)ms.zoom > ZOOMLEVEL3)
		zoom_level = 3;
	else if ((int)ms.zoom > ZOOMLEVEL2)
		zoom_level = 2;
	else 
		zoom_level = 1;

	set_scaling(&trace_scale_x,&trace_scale_y);
	output_header();


	fflush(stdout);
	out("/Times-Roman findfont");
	post_scale_font();
	out("setfont");

	for (i=1;i<zoom_level+1;i++)
	{
		sprintf(filename, "cities.file.%d", i);
		if ((cf = fopen(filename, "r")) == NULL)
			continue;

		while (read_city( cf, &city))
		{
			/* write a city name to the PS file */
			if (draw_city(&city, coord, head, zoom_level) == False)
				continue;
			post_move(coord[0], coord[1]);
			pscharstr(city.name);
			/* initialise linked list */
			if (head == NULL) {
				head = (pcity_list)malloc(sizeof(city_list));
				head->next = NULL;
				head->city = city;
			}
			/* move city into list */
			else {
				next = head;
				while (next != NULL) {
					prev = next;
					next = next->next;
				}
				next = (pcity_list)malloc(sizeof(city_list));
				next->city = city ;
				next->next = NULL;
				prev->next = next;
			}
		}
		fclose(cf);
	}
	swapbuffers();
	out("S");
	fflush(psf);
	close(psf);
}

/*****************************************************************************
the next section deals with dumping a trace to the PS file 
*****************************************************************************/
/*
Draw a filled circle at a site in the PS file
*/

void
draw_site_circle_ps(coord)
        float coord[2];
{
        float circle_size;

        circle_size = ((float) (ms.x.max_vis - ms.x.min_vis)) / 300.0;
        circle_size += circle_size * 0.5 * ms.detail;

        /* the cirlce colour is red - shows up fine in mono */
	post_arc(coord[0], coord[1], circle_size);
	out("fill");
}

/*
Draws a link between two hosts to the display
*/
void
display_link_ps(lcoord, ccoord, colour, chost, nhost)
	float lcoord[2], ccoord[2];
	int colour;
	Host_Struct *chost, *nhost;
{
	int region_in_deg;
        float rot_offset_deg;
        float deg_in_reg;

	ps_setrgbcolor(250, 0, 0); /*red*/

	/* 3d stuff - values are converted, draw arc */
        if ((ms.includes & INCL_3D) == INCL_3D)
        {
                region_in_deg = 360 / ms.side_size;
                rot_offset_deg = (float) region_in_deg * (float) ms.x_reg_offset
;
                deg_in_reg = (( (float) ms.x_offset / (float) ms.region_range) *
 (float) region_in_deg);

                rot_offset_deg = rot_offset_deg + deg_in_reg;
                rot_offset_deg = rot_offset_deg - 180.0;


                lcoord[1] = nhost->northing/3600;
                lcoord[0] = (-(nhost->easting/3600)) - rot_offset_deg;

                ccoord[1] = chost->northing/3600;
                ccoord[0] = (-(chost->easting/3600)) - rot_offset_deg;

                ThreeD_line_ps(ccoord, lcoord);
        }


	else
	{
		post_move(lcoord[0], lcoord[1]);
		post_draw(ccoord[0], ccoord[1]);
	}
	out("S");
}



/*
Passed a host and the trace parameters, outputs a host to the display
*/	
void
display_host_ps(coord, host, trace, rotation, number, sig_size)
	float coord[2];
	Host_Struct *host;
	Trace_Struct *trace;
	float rotation;
	int number;
	char sig_size;
{
	char label[170];
	char position;
	float fontx, fonty;
	double yoffset, xoffset;
	int yoff;
	struct hostent *host_ent;
	unsigned long ad;

	font_size_x(&fontx);
	font_size_y(&fonty);
if (!batchmode)
	htextsize(fontx, fonty);

	strcpy(label, "");
	if (trace->label_position == True) {
		/**label = itoa(number);*/
		/*strcat (label, ": ");*/
	}
	/*else*/
		/*strcpy(label,"");	*/

	if ((trace->label_ipnum == False) && (trace->label_ipname == False))
		strcat(label, ".");

	if (trace->label_ipnum == True) {
		if (trace->label_position == True)
			strcat(label, ": ");
	
		strcat(label, host->ipnum);
	}

	if (trace->label_ipname == True) {
		if (trace->label_ipnum == True)
			strcat(label, ", ");
		else if (trace->label_position == True)
			strcat(label, ": ");

			strcat(label, host->ipname);
	}
	
	out("gsave");

	out("/Times-Roman findfont");
	post_scale_font();
        out("setfont");


	post_trans(coord[0], coord[1]);
	
	if ((rotation > 90.0) && (rotation <= 270.0)) {
		fprintf (psf,"%lf rotate\n", rotation - 180.0);
		fprintf (psf,"(%s)\n", label);
		fprintf (psf,"stringwidth\n");
		fprintf (psf,"2.0 div\n");
		fprintf (psf,"2 sub\n");
		fprintf (psf,"round\n");
		fprintf (psf,"exch\n");
		fprintf (psf,"10 add\n");
		fprintf (psf,"neg\n");
		fprintf (psf,"exch\n");
		fprintf (psf,"translate\n");
		fprintf (psf,"0 0 moveto\n");
	}
	else
	{
		fprintf(psf,"%3.2f rotate\n",rotation);
		fprintf(psf, "(%s)\n", label);
                fprintf(psf, "stringwidth\n");
                fprintf(psf, "2.0 div\n");
                fprintf(psf, "2 sub\n");
                fprintf(psf, "round\n");
                fprintf(psf, "10\n");
                fprintf(psf, "exch\n");
                fprintf(psf, "moveto\n");
	}
	
	/* trace colour is blue - shows up fine in mono */
	ps_setrgbcolor(0, 0, 255); /*blue*/
	
	pscharstr(label);
	
	out("grestore");
}


/*
Displays a site of hosts to the screen
*/
void
display_site_ps(trace, first, last, coord)
	Trace_Struct *trace;
	int first;
	int last;
	float coord[2];
{
	Host_Struct *chost;
	int h;
	int numhosts;
	float rot;
	float current_rot = 0.0;

	/* Put a dot at this site*/
	ps_setrgbcolor(255, 0, 0);	/*red*/
	draw_site_circle_ps(coord);

	numhosts = last - first + 1; /* How many labels are to be drawn around the circle */

	rot = 360.0 / (numhosts + 1); /* Add one to the number of hosts allow for a city */

	if (rot > 36.0)
		rot = 36.0;

	current_rot = rot;

	for (h = first; h <= last; h++) {
		chost = &(trace->hosts[h]);
		display_host_ps(coord, chost, trace, current_rot, (h + 1));
		current_rot += rot;
	}
}

/*
Outputs a trace to the screen
*/
void
display_trace_ps(trace)
	Trace_Struct *trace;
{
	int h;
	Host_Struct *chost, *nhost;
	float ccoord[2], ncoord[2];
	float rotation = 0.0;
	int first = 0;
	int last = 0;

	for(h=0; h < trace->numhosts; h++) 
	{
		chost = &(trace->hosts[h]);
		if ((chost->last_at_site == True) || (h == trace->numhosts))
		{
			last = h;
			if (get_host_display_coord(chost->easting, chost->northing, ccoord))
			{
				display_site_ps(trace, first, last, ccoord);
			}

			if (h < (trace->numhosts - 1)) 
			{
				nhost = &(trace->hosts[h + 1]);
				if (get_host_display_coord(nhost->easting, nhost->northing, ncoord))
				{
						display_link_ps(ccoord, ncoord, Map_Color_Trace, chost, nhost);
				}
				else 
				{
					if (get_host_display_coord(chost->easting, chost->northing, ccoord))
						display_link_ps(ccoord, ncoord, Map_Color_Trace, chost, nhost);
				}
				first = (last + 1);
			}
		}
	}
}






/*
Outputs all the visible traces to the screen
*/
void
display_traces_ps()
{
	int t; /* Loop counters */

	set_scaling(&trace_scale_x,&trace_scale_y);
	output_header();

	/* Display each trace that is visible*/
  	for (t=0; t < ts.numtrace; t++) {
		if (ts.visible[t] == TRUE)
			display_trace_ps(&ts.traces[t]);
	out("S");
	fflush(psf);
	close(psf);
	}
}

ThreeD_line_ps(p1,p2)
    float p1[2],p2[2];
{
    float a,b,c,d;
    float q[3];
    float ang;
    float par1,par2,height;   /* for parabolic flight */
    int xp,yp;
    int v[2];
    float coord[2];
    float x_scale,y_scale;
    int x_cent,y_cent,tmp;
    long tmp_coord[2];
    float radius;
    int first_point;

    radius = 384.0;
    x_cent = 512.0;
    y_cent = 512.0;


	/*
    printf("initial lats and longs: %f %f -> %f %f\n",p1[0],p1[1],p2[0],p2[1]);
	*/

    p1[0] = deg_to_rad_ps(p1[0]);
    p1[1] = deg_to_rad_ps(p1[1]);
    p2[0] = deg_to_rad_ps(p2[0]);
    p2[1] = deg_to_rad_ps(p2[1]);

	/*
    printf("initial lats and longs: %f %f -> %f %f\n",p1[0],p1[1],p2[0],p2[1]);
	*/
 

    determine_transformation(p1,p2,&a,&b,&c,&d,radius);

	/*
    printf("rotation angles: %f %f %f\n",a,b,c,d);
	*/
    

    /*
       determine parameters of a parabola: y=ax^2+h
       where h is the height of the peak
    */

    par2 = radius / 5.0;
    par1 = - 4.0 * par2 / sqr(d); 
	
    first_point = 1;
   for(ang=0.0;ang<=d;ang+=d/40.0){
	height = sqr(ang - d/2.0) * par1 + par2;
	q[0] = 0.0;
	q[1] = (radius + height) * cos(ang);
	q[2] = (radius + height) * sin(ang);
	rotate_x_axis(q,c);
	rotate_z_axis(q,b);
	rotate_y_axis(q,a);

    	/* simply plot the x and y coords as a point after scaling */
    	/* scaling */

    	coord[0] = q[0] + x_cent;
    	coord[1] = q[1] + y_cent;


	/* if is is not on the back and inside the circle */ 
	/* this is the only case whn not to draw the line*/

	if (!((sqr(q[0]) + sqr(q[1]) < sqr(radius)) && (q[2] > 0.0)))
	{
		if (first_point == 1)
		{
			first_point = 0;
			post_move(coord[0],coord[1]);
		}
		else
			post_draw(coord[0],coord[1]);
	}
	
    }

}

