#include "geotraceman.h"
#include <sys/wait.h>
#include <stdio.h>
#include <math.h>

float trunc_secs();
void check_site();
extern Boolean convert_values();

/*
Draw a filled circle at a site
*/
void
draw_site_circle(coord)
	float coord[2];
{
	float circle_size;

	circle_size = ((float) (ms.x.max_vis - ms.x.min_vis)) / 300.0;
	circle_size += circle_size * 0.5 * ms.detail;

	/* the cirlce colour is red - shows up fine in mono */
	color(Map_Color_Dot);
	circf(coord[0], coord[1], circle_size);
}

/*
Converts a northing and an easting to appropriate coordinates to be displayed
*/
Boolean
get_host_display_coord(easting, northing, coord)
	float easting;
	float northing;
	float coord[2];
{
	float max_range = (float) ms.max_range;
	int tmp;
	long tmp_coord[2];

	coord[0] = (((easting + WORLD_MAX_LONG) / (2.0 * WORLD_MAX_LONG)) * ms.max_range);
	coord[1] = (((northing + WORLD_MAX_LAT) / (2.0 * WORLD_MAX_LAT)) * ms.max_range);

	/* Add the x rotational offset to the x coordinate */
	coord[0] = coord[0] + ((float) ms.x_offset) + ((float) ms.region_range * ms.x_reg_offset);

	if (coord[0] < 0) {
		tmp = coord[0] / ((float) ms.max_range);
		coord[0] = coord[0] + ((float) (tmp * ms.max_range));
		coord[0] = ((float) ms.max_range) - coord[0];	
	}
	else if (coord[0] > ((float) ms.max_range)) {
		tmp = coord[0] / ((float) ms.max_range);
		coord[0] = coord[0] - ((float) (tmp * ms.max_range));
	}

	if ((ms.includes & INCL_3D) == INCL_3D)
	{
		tmp_coord[0] = (long)coord[0];
		tmp_coord[1] = (long)coord[1];
		if (convert_values(tmp_coord) == True)
		{
			coord[0] = (float)tmp_coord[0];
			coord[1] = (float)tmp_coord[1];
			return True;
		}
		else
		{
			return False;
		}
	}
	else
	{
		return True;
	}
}
	
/*
Draws a link between two hosts to the display
*/
void
display_link(lcoord, ccoord, colour, chost, nhost)
	float lcoord[2], ccoord[2];
	int colour;
	Host_Struct *chost, *nhost;
{
	int region_in_deg;
	float rot_offset_deg;
	float deg_in_reg;
	/* trace colour is yellow - shows up fine on mono */
	color(Map_Color_Trace);	
	/* 3d stuff - values are converted, draw arc */
	if ((ms.includes & INCL_3D) == INCL_3D)
	{
		region_in_deg = 360 / ms.side_size;
		rot_offset_deg = (float) region_in_deg * (float) ms.x_reg_offset;
		deg_in_reg = (( (float) ms.x_offset / (float) ms.region_range) * (float) region_in_deg);
 
		rot_offset_deg = rot_offset_deg + deg_in_reg;
		rot_offset_deg = rot_offset_deg - 180.0;
		
			
		lcoord[1] = nhost->northing/3600;
		lcoord[0] = (-(nhost->easting/3600)) - rot_offset_deg;	

		ccoord[1] = chost->northing/3600;
		ccoord[0] = (-(chost->easting/3600)) - rot_offset_deg;

		ThreeD_line(ccoord, lcoord);
	}
	else
	{
		move2(lcoord[0],lcoord[1]);
		draw2(ccoord[0],ccoord[1]);
	}
}

/*
Passed a host and the trace parameters, outputs a host to the display
*/	
void
display_host(coord, host, trace, rotation, number, sig_size)
	float coord[2];
	Host_Struct *host;
	Trace_Struct *trace;
	float rotation;
	int number;
	char sig_size;
{
	char label[170];
	char position;
	float fontx, fonty;
	double yoffset, xoffset;
	int yoff;
	struct hostent *host_ent;
	unsigned long ad;

	hfont("futura.l");
	font_size_x(&fontx);
	font_size_y(&fonty);
	htextsize(fontx, fonty);
	fflush(stdout);

	strcpy(label, "");
	if (trace->label_position == True) {
		/**label = itoa(number);*/
		/*strcat (label, ": ");*/
	}
	/*else*/
		/*strcpy(label,"");	*/

	if ((trace->label_ipnum == False) && (trace->label_ipname == False))
		strcat(label, ".");

	if (trace->label_ipnum == True) {
		if (trace->label_position == True)
			strcat(label, ": ");
	
		strcat(label, host->ipnum);
	}

	if (trace->label_ipname == True) {
		if (trace->label_ipnum == True)
			strcat(label, ", ");
		else if (trace->label_position == True)
			strcat(label, ": ");

			strcat(label, host->ipname);
	}

	pushmatrix();

	hfont("futura.l");
	font_size_x(&fontx);
	font_size_y(&fonty);
	htextsize(fontx, fonty);
	yoffset =  - fonty / 2;
	xoffset = fontx;
	if ((rotation > 90.0) && (rotation <= 270.0)) {
		xoffset = - hgetstrwidth(label) - fontx;
		rotation = rotation - 180.0;
	}
	translate(coord[0], coord[1], 0.0);
	rot(rotation, 'z');
	move2(xoffset, yoffset);

	/* trace colour is yellow - shows up fine in mono */
	color(Map_Color_Txt);
	hcharstr(label);	

	popmatrix();
}


/*
Displays a site of hosts to the screen
*/
void
display_site(trace, first, last, coord)
	Trace_Struct *trace;
	int first;
	int last;
	float coord[2];
{
	Host_Struct *chost;
	int h;
	int numhosts;
	float rot;
	float current_rot = 0.0;

	/* Put a dot at this site*/
	draw_site_circle(coord);

	numhosts = last - first + 1; /* How many labels are to be drawn around the circle */

	rot = 360.0 / (numhosts + 1); /* Add one to the number of hosts allow for a city */

	if (rot > 36.0)
		rot = 36.0;

	current_rot = rot;

	for (h = first; h <= last; h++) {
		chost = &(trace->hosts[h]);
		display_host(coord, chost, trace, current_rot, (h + 1));
		current_rot += rot;
	}
}

/*
Outputs a trace to the screen
*/
void
display_trace(trace)
	Trace_Struct *trace;
{
	int h;
	Host_Struct *chost, *nhost;
	float ccoord[2], ncoord[2];
	float rotation = 0.0;
	int first = 0;
	int last = 0;

	for(h=0; h < trace->numhosts; h++) 
	{
		chost = &(trace->hosts[h]);
		if ((chost->last_at_site == True) || (h == trace->numhosts))
		{
			last = h;
			if (get_host_display_coord(chost->easting, chost->northing, ccoord))
			{
				display_site(trace, first, last, ccoord);
			}

			if (h < (trace->numhosts - 1)) 
			{
				nhost = &(trace->hosts[h + 1]);
				if (get_host_display_coord(nhost->easting, nhost->northing, ncoord))
				{
if (!batchmode)
						display_link(ccoord, ncoord, Map_Color_Trace, chost, nhost);
				}
				else 
				{
					if (get_host_display_coord(chost->easting, chost->northing, ccoord))
if (!batchmode)
						display_link(ccoord, ncoord, Map_Color_Trace, chost, nhost);
				}
				first = (last + 1);
			}
		}
	}
}

/*
Outputs all the visible traces to the screen
*/
void
display_traces()
{
	int t; /* Loop counters */

	/* Display each trace that is visible*/
  	for (t=0; t < ts.numtrace; t++) {
		if (ts.visible[t] == TRUE)
			display_trace(&ts.traces[t]);
	}
}

/*
Outputs the last read in host to the display
*/
void
display_extra_host(trace, first_host, sig_size)
	Trace_Struct *trace;
	Boolean first_host;
	char sig_size;
{
	float tmp;
	float lcoord[2], ccoord[2];
	Host_Struct *chost, *lhost;
	static float rotation = 36.0;

	chost = &((*trace).hosts[(trace->numhosts - 1)]);
	if (!get_host_display_coord(chost->easting, chost->northing, ccoord))
		{
		if (!((ms.includes & INCL_3D) == INCL_3D))
			return;
		}

	check_site(ccoord);
	if (first_host == FALSE) {
		lhost = &((*trace).hosts[(trace->numhosts - 2)]);
		if (!get_host_display_coord(lhost->easting, lhost->northing, lcoord))
			{
			if (!((ms.includes & INCL_3D) == INCL_3D))
				return;
			}

		check_site(lcoord);
		if ((lcoord[0] == ccoord[0]) && (lcoord[1] == ccoord[1])) {
			rotation += 36.0;
			lhost->last_at_site = False;
		}
		else 
		{
if (!batchmode)
{
			display_link(lcoord, ccoord, Map_Color_Trace, chost, lhost);
                        draw_site_circle(ccoord);
}
                       	rotation = 36.0;
                       	lhost->last_at_site = True;
		}
	}
	if (get_host_display_coord(chost->easting, chost->northing, ccoord))
	{
if(!batchmode)
{
		display_host(ccoord, chost, trace, rotation, trace->numhosts, sig_size);
		draw_site_circle(ccoord);
}
	}
}


void check_site(coord)
float *coord;
{
	static phost_list head;
	phost_list tmp, prev;
	Boolean in_list = False;

	if (head == NULL) {
		head = (phost_list) malloc(sizeof(host_list));
		head->next = NULL;
		head->northing = coord[1];
		head->easting = coord[0];
		return;
	}

	tmp = prev = head;

	while (tmp != NULL) {
		if ((coord[0] == tmp->easting) && (coord[1] == tmp->northing))
			in_list = True;
		if ((int)ms.zoom >= ZOOMLEVEL2) {
			/* strip off seconds */
			if (trunc_secs(coord[1]) == trunc_secs(tmp->northing))
				coord[1] = tmp->northing;
			if (trunc_secs(coord[0]) == trunc_secs(tmp->easting))
				coord[0] = tmp->easting;
		}
		else if (((int)ms.zoom < ZOOMLEVEL2) && ((int)ms.zoom >= ZOOMLEVEL1)) {
			/* strip off minutes and seconds */
			if ((int)coord[1] == (int)tmp->northing)
				coord[1] = tmp->northing;
			if ((int)coord[0] == (int)tmp->easting)
				coord[0] = tmp->easting;
		}
		if (in_list == True)
			return;
		prev = tmp;
		tmp = tmp->next;
	}
	tmp = (phost_list) malloc(sizeof(host_list));
	tmp->next = NULL;
	tmp->northing = coord[1];
	tmp->easting = coord[0];
	prev->next = tmp;
}

float trunc_secs(value)
float value;
{
	float return_val;
	float tmp1, tmp2;

	/* strip off degrees */
	tmp1 = value - (int)value;
	return_val = tmp1;
	
	/* strip off minutes */
	tmp1 *= 60;
	tmp2 = tmp1 - (int)tmp1;
	return_val += tmp2;

	/* return the value without the seconds */
	return return_val;
}

